




"""
pyscal module containing methods for processing of a trajectory. Methods for
reading of input files formats, writing of output files etc are provided in
this module.

"""
import numpy as np
import gzip
import pyscal.catom as pca
from ase import Atom, Atoms
import gzip
import io
import os
import warnings
from ase.io import write, read
import pyscal.formats.ase as ptase
import pyscal.formats.lammps as ptlammps
import pyscal.formats.mdtraj as ptmdtraj
import pyscal.formats.vasp as ptvasp


def read_file(filename, format="lammps-dump",
    compressed = False, customkeys=None):
    """
    Read input file

    Parameters
    ----------
    filename : string
        name of the input file.

    format : {'lammps-dump', 'poscar', 'ase', 'mdtraj'}
        format of the input file, in case of `ase` the ASE Atoms object

    compressed : bool, optional
        If True, force to read a `gz` compressed format, default False.

    customkeys : list
        A list containing names of headers of extra data that needs to be read in from the
        input file.

    Returns
    -------
    None
    """

    if customkeys == None:
        customkeys = []
    customread = (len(customkeys) > 0)

    if format=='lammps-dump':
        atoms, box, is_triclinic = ptlammps.read_snap(filename, compressed=compressed, check_triclinic=True, customkeys=customkeys)
    elif format == 'ase':
        atoms, box = ptase.read_snap(filename)
    elif format == 'mdtraj':
        atoms, box = ptmdtraj.read_snap(filename)
    elif format == 'poscar':
        atoms, box = ptvasp.read_snap(filename, compressed=compressed)
    else:
        try:
            #try to use ASE backend
            aseobject = read(filename, format=format)
            atoms, box = ptase.read_snap(aseobject)
            warnings.info("Using ase backend to read file")
        except:
            raise TypeError("format recieved an unknown option %s"%format)

    return atoms, box       

def write_file(sys, outfile, format="lammps-dump", compressed = False, 
    customkeys=None, customvals=None, timestep=0, species=None):
    """
    Write the state of the system to a trajectory file.

    Parameters
    ----------
    sys : `System` object
        the system object to be written out

    outfile : string
        name of the output file

    format : string, optional
        format of the output file

    compressed : bool, default false
        write a `.gz` format

    customkey : string or list of strings, optional
        If specified, it adds this custom column to the dump file. Default None.

    customvals : list or list of lists, optional
        If `customkey` is specified, `customvals` take an array of the same length
        as number of atoms, which contains the values to be written out.
    
    timestep: int, optional
        Specify the timestep value, default 0

    species : None, optional
        species of the atoms. Required if any format other than 'lammps-dump' is used. Required
        for convertion to ase object.


    Returns
    -------
    None

    """
    if format=='lammps-dump':
        ptlammps.write_snap(sys, outfile, compressed = compressed,
        customkeys = customkeys, customvals = customvals, timestep = timestep)
    
    elif format == 'lammps-data':
        aseobject = ptase.convert_snap(sys, species=species)
        write(outfile, aseobject, format='lammps-data')
    
    elif format == 'poscar':
        ptvasp.write_snap(sys, outfile, species=species)
        #aseobject = ptase.convert_snap(sys, species=species)
        #write(outfile, aseobject, format='vasp')

    else:
        #try a write using ase
        try:
            aseobject = ptase.convert_snap(sys, species=species)
            write(outfile, aseobject, format=format)
            warnings.info("Using ase backend to write file")
        except:
            raise TypeError("format recieved an unknown option %s"%format)


def split_trajectory(infile, format='lammps-dump', compressed=False):
    """
    Read in a trajectory file and convert it to individual time slices.

    Parameters
    ----------

    filename : string
        name of input file

    format : format of the input file
        only `lammps-dump` is supported now.

    compressed : bool, optional
        force to read a `gz` zipped file. If the filename ends with `.gz`, use of this keyword is not
        necessary.

    Returns
    -------
    snaps : list of strings
        a list of filenames which contain individual frames from the main trajectory.

    Notes
    -----
    This is a wrapper function around `split_traj_lammps_dump` function.

    """
    if not os.path.exists(infile):
        raise FileNotFoundError("Filename %s not found"%infile)

    snaps = []

    if format=='lammps-dump':
        snaps = ptlammps.split_snaps(infile, compressed = compressed)
    elif format == 'ase':
        snaps = ptase.split_snaps()
    elif format == 'mdtraj':
        snaps = ptmdtraj.split_snaps()
    elif format == 'poscar':
        snaps = ptvasp.split_snaps()
    else:
        raise TypeError("format recieved an unknown option %s"%format)

    return snaps





import pyscal.core as pc
import numpy as np
import warnings

def compare_atomic_env(infile, atomtype=2, precision=2, format="poscar", print_results=True, return_system=False):
    """
    Compare the atomic environment of given types of atoms
    in the inputfile. The comparison is made in terms of Voronoi
    volume and Voronoi fingerprint.

    Parameters
    ----------
    infile : string
        name of the inputfile

    atomtype: int, optional
        type of the atom
        default 2

    precision: float, optional
        precision for comparing Voronoi volumes
        default 3

    format: string, optional
        format of the input file
        default poscar

    print_results: bool, optional
        if True, print the results. If False, return the data
        instead. default True

    return_system: bool, optional
        if True, return the system object.
        default False

    Returns
    -------
    vvx : list of floats
        unique Voronoi volumes. Returned only if print results is False

    vrx : list of strings
        unique Voronoi polyhedra. Returned only if print results is False

    vvc : list of ints
        number of unique quantities specified above. Returned only if print results is False
    """
    sys = pc.System()
    sys.read_inputfile(infile, format=format)
    sys.find_neighbors(method="voronoi")
    sys.calculate_vorovector()
    atoms = sys.atoms
    vols = []
    vors = []
    for atom in atoms:
        if atom.type == atomtype:
            vols.append(atom.volume)
            vors.append(" ".join(np.array(atom.vorovector).astype(str)))
    vols = np.array(vols)
    vols = np.round(vols, decimals=precision)
    vvx, vvc = np.unique(vols, return_counts=True)
    vrx, vrc = np.unique(vors, return_counts=True)

    length_mismatch = False
    if (len(vvx) != len(vrx)):
        warnings.warn("Different voronoi polyhedra with same volume! Fingerprint wont be printed. Maybe change precision?")
        length_mismatch = True

    if precision > 3:
        warnings.warn("More than 3 digits of precision selected!")

    if print_results:
        print("%d clusters found"%len(vvx))
        if not length_mismatch:
            for i in range(len(vvx)):
                print("voro fingerprint = <%s>, vol = %.3f, counts = %d"%(vrx[i], vvx[i], vvc[i]))
        else:
            for i in range(len(vvx)):
                print("voro fingerprint = <x x x x>, vol = %.3f, counts = %d"%(vvx[i], vvc[i]))
        if return_system:
            return sys
    else:
        if return_system:
            return np.array([vvx, vrx, vvc]), sys
        else:
            return np.array([vvx, vrx, vvc])


def find_tetrahedral_voids(infile, format="poscar", print_results=True, return_system=False, 
                                direct_coordinates=True, precision=0.1):
    """
    Check for tetrahedral voids in the system

    Parameters
    ----------
    infile : string
        name of the input file

    format : string
        format of the input file, optional
        default poscar

    print_results: bool, optional
        if True, print the results. If False, return the data
        instead. default True

    return_system: bool, optional
        if True, return the system object.
        default False

    direct_coordinates: bool, optional
        if True, results are provided in direct coordinates
        default False

    precision: int, optional
        the number of digits to check for distances.
        default 1
    
    Returns
    -------
    types : list of atom types
    volumes : list of atom volumes
    pos : list of atom positions
    sys : system object, returns only if return_sys is True

    """

    sys = pc.System()
    sys.read_inputfile(infile, format=format)
    sys.find_neighbors(method="voronoi")
    sys.calculate_vorovector()
    atoms = sys.atoms

    volumes = []
    pos = []
    types = []

    box = sys.box
    boxx = box[0][1] - box[0][0]
    boxy = box[1][1] - box[1][0]
    boxz = box[2][1] - box[2][0]

    if len(atoms) < 21:
        warnings.warn("Very less number of atoms, results maybe wrong. See https://github.com/srmnitc/pyscal/issues/63 ")

    for atom in atoms:
        if atom.vorovector[1] == 4:
            #first to prevent small cells we need to filter
            neighs = np.unique(atom.neighbors)
            dists = np.array([sys.get_distance(atom, atoms[x]) for x in neighs])
            mindist = dists[np.argsort(dists)][0]
            mindistcount = 0
            for i in range(len(dists)):
                if mindist*(1.0-precision) <= dists[i] <= mindist*(1.0+precision):
                    mindistcount += 1

            if mindistcount == 4:
                volumes.append(atom.volume)
                if direct_coordinates:
                    p = atom.pos
                    pos.append([p[0]/boxx, p[1]/boxy, p[2]/boxz])
                else:
                    pos.append(atom.pos)
                types.append(atom.type)

    if print_results:
        if len(volumes) > 0:
            print("%d atoms found in tetrahedral position"%len(volumes))
            for i in range(len(volumes)):
                print("%d  volume %f type %d at %f, %f, %f"%(i+1, volumes[i], types[i], pos[i][0], pos[i][1], pos[i][2]))
        else:
            print("no atoms found in tetrahedral position")

    if return_system:
        return types, volumes, pos, sys
    else:
        return types, volumes, pos


import numpy as np
import gzip
import pyscal.catom as pca
from ase import Atom, Atoms
import gzip
import io
import os

#new function to wrap over mdtraj objects
def read_snap(mdobject, check_triclinic=False):
    """
    Function to read from an MDTraj atoms objects

    Parameters
    ----------
    mdobject : MDTraj Atoms object
        name of the MDTraj atoms object

    triclinic : bool, optional
        True if the configuration is triclinic

    """
    #We have to process atoms and atomic objects from ase
    #Known issues lammps -dump modified format
    #first get box
    a = np.array(mdobject.unitcell_vectors[0][0])
    b = np.array(mdobject.unitcell_vectors[0][1])
    c = np.array(mdobject.unitcell_vectors[0][2])

    box = np.array([a, b, c])

    #box and box dims are set. Now handle atoms
    chems = np.array([atom.name for atom in mdobject.topology.atoms])
    atomsymbols = np.unique(chems)
    atomtypes = np.array(range(1, len(atomsymbols)+1))
    typedict = dict(zip(atomsymbols, atomtypes))

    #now start parsing atoms
    atoms = []
    positions = mdobject.xyz[0]
    for count, position in enumerate(positions):
        atom = pca.Atom()
        atom.pos = list(position)
        atom.id = count+1
        atom.type = typedict[chems[count]]
        atom.loc = count

        customdict = {'species': chems[count]}
        atom.custom = customdict
        atoms.append(atom)

    return atoms, box

def write_snap(**kwargs):
    raise NotImplementedError("write method for mdtraj is not implemented")

def split_snaps(**kwargs):
    raise NotImplementedError("split method for mdtraj is not implemented")

def convert_snap(**kwargs):
    raise NotImplementedError("convert method for mdtraj is not implemented")



"""
The pyscal routines module. This will contain slightly bigger routines which needs to be used.
Most of them will not be used at the user level. User level longer routines will go to misc
module.

Imports might need specific modules!
"""

import numpy as np
import os


def get_energy_atom(outfile, species=None, pair_style=None, pair_coeff=None, mass=1.0):
    """
    Get energy per atom using a LAMMPS calculator
    """

    #file is written
    #now lammps part
    try:
        from lammps import lammps
    except ImportError:
        raise ModuleNotFoundError("energy method needs lammps compiled as a library. Either install with conda - conda install -c conda-forge lammps,\
                or see here - https://lammps.sandia.gov/doc/Howto_pylammps.html")

    #start routine
    lmp = lammps()
    lmp.command("echo log")
    lmp.command("units metal")
    lmp.command("atom_style atomic")
    lmp.command("boundary p p p")

    if (pair_style is None):
        raise ValueError("pair style has to be provided")

    if (pair_coeff is None):
        raise ValueError("pair coeff has to be provided")

    lmp.command('read_data %s'%outfile)
    lmp.command('pair_style %s'%pair_style)
    lmp.command('pair_coeff %s'%pair_coeff)
    lmp.command("mass * %f"%mass)

    lmp.command("compute 1 all pe/atom")
    lmp.command("run 0")
    eng = lmp.extract_compute("1",1,1)
    ids = lmp.extract_atom("id",0)

    natoms = lmp.get_natoms()
    eng = [eng[x] for x in range(natoms)]
    ids = [ids[x] for x in range(natoms)]

    iddict = dict(zip(np.array(ids).astype(str), eng))
    return iddict





import numpy as np
import gzip
import pyscal.catom as pca
from ase import Atom, Atoms
import gzip
import io
import os

#functions that are not wrapped from C++
def read_snap(infile, compressed = False, check_triclinic=False, customkeys=None):
    """
    Function to read a lammps dump file format - single time slice.

    Parameters
    ----------
    infile : string
        name of the input file

    compressed : bool, optional
        force to read a `gz` zipped file. If the filename ends with `.gz`, use of this keyword is not
        necessary. Default True.

    check_triclinic : bool, optional
        If true check if the sim box is triclinic. Default False.

    customkeys : list of strings, optional
        A list of extra keywords to read from trajectory file.

    Returns
    -------
    atoms : list of `Atom` objects
        list of all atoms as created by user input

    boxdims : list of list of floats
        The dimensions of the box. This is of the form `[[xlo, xhi],[ylo, yhi],[zlo, zhi]]` where `lo` and `hi` are
        the upper and lower bounds of the simulation box along each axes. For triclinic boxes, this is scaled to
        `[0, scalar length of the vector]`.

    box : list of list of floats
        list of the type `[[x1, x2, x3], [y1, y2, y3], [zz1, z2, z3]]` which are the box vectors. Only returned if
        `box_vectors` is set to True.

    triclinic : bool
        True if the box is triclinic. Only returned if `check_triclinic` is set to True

    .. note::

        Values are always returned in the order `atoms, boxdims, box, triclinic` if all
        return keywords are selected. For example, ff `check_triclinic` is not selected, the return
        values would still preserve the order and fall back to  `atoms, boxdims, box`.

    Notes
    -----
    Read a lammps-dump style snapshot that can have variable headers, reads in type and so on.
    Zipped files which end with a `.gz` can also be read automatically. However, if the file does not
    end with a `.gz` extension, keyword `compressed = True` can also be used.

    Examples
    --------
    >>> atoms, box = read_lammps_dump('conf.dump')
    >>> atoms, box = read_lammps_dump('conf.dump.gz')
    >>> atoms, box = read_lammps_dump('conf.d', compressed=True)

    """
    if isinstance(infile, list):
        islist = True
        f = infile
    else:
        islist = False

    if not islist:
        if not os.path.exists(infile):
            raise FileNotFoundError("Filename %s not found"%infile)
        #first depending on if the extension is .gz - use zipped read
        raw = infile.split('.')
        if raw[-1] == 'gz' or  compressed:
            f = gzip.open(infile,'rt')
        else:
            f = open(infile,'r')

    if customkeys == None:
        customkeys = []

    #now go through the file line by line
    paramsread = False
    atoms = []
    triclinic = False
    volume_fraction = 1.00

    #now if custokeys are provided - read those in too
    customread = False
    customlength = len(customkeys)
    if customlength > 0:
        customread = True

    nblock = 0
    for count, line in enumerate(f):
        #print(count, line)
        if not paramsread:
            #atom numer is at line 3
            if count == 3:
                natoms = int(line.strip())
                nblock = natoms+9
            #box dims in lines 5,6,7
            elif count == 5:
                raw = line.strip().split()
                boxx = [float(raw[0]), float(raw[1])]
                if len(raw) == 3:
                    xy = float(raw[2])
            elif count == 6:
                raw = line.strip().split()
                boxy = [float(raw[0]), float(raw[1])]
                if len(raw) == 3:
                    xz = float(raw[2])
            elif count == 7:
                raw = line.strip().split()
                boxz = [float(raw[0]), float(raw[1])]
                if len(raw) == 3:
                    yz = float(raw[2])
                    triclinic = True
                    tilts = [xy, xz, yz]
                #boxdims = [boxx, boxy, boxz]

            #header is here
            elif count == 8:
                raw = line.strip().split()
                headerdict = { raw[x]:x-2 for x in range(0, len(raw)) }
                paramsread = True
                if customread:
                    if not np.prod([(x in headerdict) for x in customkeys]):
                        raise KeyError("some values in custokeys was not found in the file")
                #add a new keyword for scaled coordinates
                if "x" in headerdict.keys():
                    scaled = False
                elif "xs" in headerdict.keys():
                    scaled = True
                    headerdict["x"] = headerdict.pop("xs")
                    headerdict["y"] = headerdict.pop("ys")
                    headerdict["z"] = headerdict.pop("zs")
                else:
                    raise ValueError("only x/xs, y/ys andz/zs keys are allowed for traj file")


        else:
            if count == nblock:
                break
            raw = line.strip().split()
            idd = int(raw[headerdict["id"]])
            typ = int(raw[headerdict["type"]])
            x = float(raw[headerdict["x"]])
            y = float(raw[headerdict["y"]])
            z = float(raw[headerdict["z"]])

            atom = pca.Atom()
            atom.pos = [x, y, z]
            atom.id = idd
            atom.type = typ
            atom.loc = count-8

            customdict = {}
            #if customkeys need to be read, do it
            if customread:
                for cc, kk in enumerate(customkeys):
                    customdict[kk] = raw[headerdict[kk]]

            atom.custom = customdict
            atoms.append(atom)

    #close files
    if not islist:
        f.close()

    if triclinic:
        #process triclinic box
        amin = min([0.0, tilts[0], tilts[1] ,tilts[0]+tilts[1]])
        amax = max([0.0, tilts[0], tilts[1] ,tilts[0]+tilts[1]])
        bmin = min([0.0, tilts[2]])
        bmax = max([0.0, tilts[2]])
        xlo = boxx[0] - amin
        xhi = boxx[1] - amax
        ylo = boxy[0] - bmin
        yhi = boxy[1] - bmax
        zlo = boxz[0]
        zhi = boxz[1]

        #triclinic cell
        a = np.array([xhi-xlo, 0, 0])
        b = np.array([tilts[0], yhi-ylo, 0])
        c = np.array([tilts[1], tilts[2], zhi-zlo])

        rot = np.array([a, b, c]).T
        rotinv = np.linalg.inv(rot)
        ortho_origin = np.array([boxx[0], boxy[0], boxz[0]])

        for atom in atoms:
            #correct zero of the atomic positions (shift box to origin)
            dist = np.array(atom.pos) - ortho_origin
            atom.pos = dist

        #finally change boxdims - to triclinic box size
        box = np.array([a, b, c])
    else:
        box = np.array([[boxx[1]-boxx[0], 0, 0],[0, boxy[1]-boxy[0], 0],[0, 0, boxz[1]-boxz[0]]])

    #adjust for scled coordinates
    if scaled:
        for atom in atoms:
            dist = atom.pos
            ndist = dist[0]*box[0] + dist[1]*box[1] + dist[2]*box[2]
            atom.pos = ndist

    if check_triclinic:
        return atoms, box, triclinic
    else:
        return atoms, box

def write_snap(sys, outfile, compressed = False, 
    customkeys=None, customvals=None, timestep=0):
    """
    Write the state of the system to a trajectory file.

    Parameters
    ----------
    sys : `System` object
        the system object to be written out

    outfile : string
        name of the output file

    compressed : bool, default false
        write a `.gz` format

    customkey : string or list of strings, optional
        If specified, it adds this custom column to the dump file. Default None.

    customvals : list or list of lists, optional
        If `customkey` is specified, `customvals` take an array of the same length
        as number of atoms, which contains the values to be written out.
        shape: natoms x ncustomkeys
    
    timestep: int, optional
        Specify the timestep value, default 0

    Returns
    -------
    None

    """
    if customkeys == None:
        customkeys = []

    box = sys.box
    boxx = np.sqrt(np.sum(np.array(box[0])**2))
    boxy = np.sqrt(np.sum(np.array(box[1])**2))
    boxz = np.sqrt(np.sum(np.array(box[2])**2))

    atoms = sys.atoms

    if len(customkeys) > 0:
        if customvals is None:
            cvals = [sys.get_custom(atom, customkeys) for atom in atoms]
        else:
            #first check if dim is equal to keys dim
            shape = np.array(customvals).shape
            rqdshape = (len(atoms), len(customkeys))
            if shape != rqdshape:
                raise ValueError("Customvals should be of shape natoms x ncustomkeys. Found %d-%d, should be %d-%d"%(shape[0], 
                    shape[1], rqdshape[0], rqdshape[1]))
            cvals = customvals

    #open files for writing
    if isinstance(outfile, io.IOBase):
        dump = outfile
    else:
        if compressed:
            gz = gzip.open(outfile,'wt')
            dump = gz
        else:
            gz = open(outfile,'w')
            dump = gz

    #now write
    dump.write("ITEM: TIMESTEP\n")
    dump.write("%d\n" % timestep)
    dump.write("ITEM: NUMBER OF ATOMS\n")
    dump.write("%d\n" % len(atoms))
    dump.write("ITEM: BOX BOUNDS\n")
    dump.write("%f %f\n" % (0, boxx))
    dump.write("%f %f\n" % (0, boxy))
    dump.write("%f %f\n" % (0, boxz))

    #now write header
    if len(customkeys) > 0:
        ckey = " ".join(customkeys)
        title_str = "ITEM: ATOMS id type x y z %s\n"% ckey
    else:
        title_str = "ITEM: ATOMS id type x y z\n"

    dump.write(title_str)

    for cc, atom in enumerate(atoms):
        pos = atom.pos
        if len(customkeys) > 0:
            cval_atom = " ".join(np.array(list(cvals[cc])).astype(str))
            atomline = ("%d %d %f %f %f %s\n")%(atom.id, atom.type, pos[0], pos[1], pos[2], cval_atom)
        else:
            atomline = ("%d %d %f %f %f\n")%(atom.id, atom.type, pos[0], pos[1], pos[2])

        dump.write(atomline)

    if not isinstance(outfile, io.IOBase):
        dump.close()


def split_snaps(infile, compressed = False):
    """
    Read in a LAMMPS dump trajectory file and convert it to individual time slices.

    Parameters
    ----------

    filename : string
        name of input file

    compressed : bool, optional
        force to read a `gz` zipped file. If the filename ends with `.gz`, use of this keyword is not
        necessary, Default False.

    Returns
    -------
    snaps : list of strings
        a list of filenames which contain individual frames from the main trajectory.

    """
    raw = infile.split('.')
    if raw[-1] == 'gz' or  compressed:
        f = gzip.open(infile,'rt')
    else:
        f = open(infile,'r')


    #pre-read to find the number of atoms
    for count, line in enumerate(f):
            if count == 3:
                natoms = int(line.strip())
                break
    f.close()

    #now restart f()
    if raw[-1] == 'gz' or  compressed:
        f = gzip.open(infile,'rt')
    else:
        f = open(infile,'r')


    nblock = natoms+9
    startblock = 0
    count=1
    snaps = []



    for line in f:
        if(count==1):
            ff = ".".join([infile, 'snap', str(startblock), 'dat'])
            lines = []
            lines.append(line)

        elif(count<nblock):
            lines.append(line)

        else:
            lines.append(line)
            snaps.append(ff)
            fout = open(ff,'w')
            for wline in lines:
                fout.write(wline)
            fout.close()

            count=0
            startblock+=1
        count+=1

    f.close()

    return snaps

def convert_snap(**kwargs):
    raise NotImplementedError("convert method for mdtraj is not implemented")
